#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['modbus_server']

package_data = \
{'': ['*']}

setup(name='modbus_server',
      version='0.1.0',
      description='Easy-to-use ModbusTCP Server in pure Python ',
      author='Christoph Kattmann',
      author_email='christoph.kattmann@bss-hs.de',
      url=None,
      packages=packages,
      package_data=package_data,
     )
